from datetime import datetime
from typing import List, Tuple
from threading import Lock


class CloudWatchStreamWriter(object):

    def __init__(self, logs_client, group_name: str, stream_name: str):
        self.logs_client = logs_client
        self.group_name = group_name
        self.stream_name = stream_name
        self._sequence_token = None
        self._lock = Lock()
        self._init_stream()

    def _init_stream(self):
        try:
            # check if logging group already exist
            log_groups = self.logs_client.describe_log_groups(logGroupNamePrefix=self.group_name)
            if len(log_groups['logGroups']) == 0:
                # if logging group not exists create it and set retention for 30 days
                self.logs_client.create_log_group(logGroupName=self.group_name)
                self.logs_client.put_retention_policy(logGroupName=self.group_name, retentionInDays=30)

            # check if stream already exist
            log_streams = self.logs_client.describe_log_streams(logGroupName=self.group_name,
                                                                logStreamNamePrefix=self.stream_name)
            if len(log_streams['logStreams']) == 0:
                # if not exists create the stream
                response = self.logs_client.create_log_stream(logGroupName=self.group_name,
                                                              logStreamName=self.stream_name)

                log_streams = self.logs_client.describe_log_streams(logGroupName=self.group_name,
                                                                    logStreamNamePrefix=self.stream_name)

            log_stream = log_streams['logStreams'][0]
            self._sequence_token = log_stream.get('uploadSequenceToken') or "0"
        except Exception as ex:
            print(ex)
            raise

    @staticmethod
    def _fix_empty_string(string: str) -> str:
        if len(string) == 0:
            return " "
        return string

    def write(self, log_events: List[Tuple[datetime, str]]):
        """Write log events to AWS CloudWatch Logs service

        Args:
            log_events: list of tuples containing timestamp and log message
        """
        try:
            events = [
                {
                    'timestamp': int(time.timestamp() * 1000),
                    'message': self._fix_empty_string(message)
                } for time, message in log_events]

            with self._lock:
                result = self.logs_client.put_log_events(logGroupName=self.group_name,
                                                         logStreamName=self.stream_name,
                                                         logEvents=events,
                                                         sequenceToken=self._sequence_token)
                if "rejectedLogEventsInfo" in result:
                    raise Exception(result["rejectedLogEventsInfo"])

                self._sequence_token = result['nextSequenceToken']

        except Exception as ex:
            print(ex)
            raise



import boto3
from typing import Dict
from threading import Lock
from ..logs import LogEntry, ICloudLogger
from .stream_writer import CloudWatchStreamWriter


class CloudWatchLogger(ICloudLogger):
    def __init__(self, config: dict):
        region = config["region_name"]
        sandbox = config["sandbox_id"]
        self.logs = boto3.client('logs', region_name=region)
        self.group_name = "/colony/sandboxes/{sandbox}".format(sandbox=sandbox)
        self._stream_writers = dict()  # type: Dict[str, CloudWatchStreamWriter]
        self._lock = Lock()

    def _get_writer(self, log_entry: LogEntry) -> CloudWatchStreamWriter:
        stream_name = "/{0}/{1}/{2}".format(log_entry.app, log_entry.instance, log_entry.topic)
        writer = self._stream_writers.get(stream_name)
        if writer is None:
            with self._lock:
                writer = self._stream_writers.get(stream_name)
                if writer is None:
                    writer = CloudWatchStreamWriter(self.logs, self.group_name, stream_name)
                self._stream_writers[stream_name] = writer
        return writer

    def write(self, log_entry: LogEntry):
        writer = self._get_writer(log_entry)
        writer.write(log_entry.log_events)



import json
import logging
import os
from typing import List

from datetime import datetime
from flask import Flask, request
from flask import jsonify

from sidecar.app_configuration_start_policy import AppConfigurationStartPolicy
from sidecar.app_instance_config_status_event_reporter import AppInstanceConfigStatusEventReporter
from sidecar.app_instance_event_handler import AppInstanceEventHandler
from sidecar.app_instance_events import AppInstanceEvents
from sidecar.app_instance_identifier_creator import IAppInstanceIdentifierCreator
from sidecar.app_request_info import AppRequestInfo
from sidecar.app_status_maintainer import AppStatusMaintainer
from sidecar.apps_configuration_end_tracker import AppsConfigurationEndTracker
from sidecar.aws_app_instance_identifier_creator import AwsAppInstanceIdentifierCreator
from sidecar.aws_app_instance_service import AwsAppInstanceService
from sidecar.aws_sandbox_start_time_updater import AwsSandboxStartTimeUpdater
from sidecar.aws_sandbox_deployment_end_updater import AwsSandboxDeploymentEndUpdater
from sidecar.aws_session import AwsSession
from sidecar.aws_vpc_service import AwsVpcService
from sidecar.cloud_logger import ICloudLogger, LogEntry
from sidecar.const import Const, DateTimeProvider
from sidecar.file_system import FileSystemService
from sidecar.health_check_monitor import HealthCheckMonitor
from sidecar.kub_api_service import KubApiService
from sidecar.kub_app_instance_identifier_creator import KubAppInstanceIdentifierCreator
from sidecar.kub_app_instance_service import KubAppInstanceService
from sidecar.kub_sandbox_start_time_updater import KubSandboxStartTimeUpdater
from sidecar.kub_sandbox_deployment_end_updater import KubSandboxDeploymentEndUpdater
from sidecar.kub_token_provider import KubTokenProvider
from sidecar.messaging_service import MessagingService, MessagingConnectionProperties
from sidecar.sandbox_deployment_state_tracker import SandboxDeploymentStateTracker
from sidecar.aws_status_maintainer import AWSStatusMaintainer

app = Flask(__name__)
logger = None  # type: logging.Logger
cloud_logger = None  # type: ICloudLogger
_health_check_monitor = None  # type: HealthCheckMonitor
_app_configuration_start_policy = None  # type: AppConfigurationStartPolicy
_app_instance_identifier_creator = None  # type: IAppInstanceIdentifierCreator
_app_instance_event_handler = None  # type: AppInstanceEventHandler



def _load_config():
    with open(Const.get_config_file(), 'r') as conf:
        config_json = json.loads(conf.read())
        return config_json


def _convert_messaging_json_to_connection_props(messaging_json: dict()) -> MessagingConnectionProperties:

    return MessagingConnectionProperties(messaging_json["host"], messaging_json["user"], messaging_json["password"],
                                         messaging_json["queue"], messaging_json["exchange"], messaging_json["routingkey"],
                                         messaging_json["virtualhost"], messaging_json["port"], messaging_json["queuetype"])


def _convert_apps_json_to_app_requests(apps_json: dict()) -> List[AppRequestInfo]:
    # TODO: remove the conditional assignment once all branches are up-to-date with issue #95
    app_requests = [AppRequestInfo(name=app_name,
                                   instances_count=app_value["instances_count"] if isinstance(app_value,
                                                                                              dict) else app_value,
                                   app_dependencies=app_value["dependencies"] if isinstance(app_value, dict) else [])
                    for app_name, app_value in apps_json.items()]
    return app_requests


def _get_apps_environment_variables(config_json: dict) -> dict:
    env = {app_name: app_config["env"] for app_name, app_config in config_json["apps"].items()}
    return env


def get_status_updater(config_json, app_instance_event_handler: AppInstanceEventHandler):
    try:
        provider = config_json["provider"]
        sandbox_id = config_json['sandbox_id']
        space_id = config_json['space_id']
        apps_json = config_json['apps']
        app_requests = _convert_apps_json_to_app_requests(apps_json=apps_json)

        messaging_props = _convert_messaging_json_to_connection_props(config_json['messaging'])

        messaging_service = MessagingService(messaging_props, logger)

        app_instance_status_event_reporter = AppInstanceConfigStatusEventReporter(
            app_instance_event_handler=app_instance_event_handler)
        if provider == "kubernetes":
            kub_token_provider = KubTokenProvider()
            hostname = config_json['kub_api_address']
            kas = KubApiService(hostname=hostname, namespace=sandbox_id, kub_token_provider=kub_token_provider)

            app_instance_service = KubAppInstanceService(logger=logger, kub_api_service=kas)
            apps_configuration_end_tracker = AppsConfigurationEndTracker(logger=logger, app_requests=app_requests,
                                                                         app_instance_service=app_instance_service)
            app_configuration_start_policy = AppConfigurationStartPolicy(apps_configuration_end_tracker,
                                                                         app_requests)

            sandbox_start_time_updater = KubSandboxStartTimeUpdater(
                date_time_provider=DateTimeProvider(),
                logger=logger,
                kub_api_service=kas,
                apps_configuration_end_tracker=apps_configuration_end_tracker)

            sandbox_deployment_end_updater = KubSandboxDeploymentEndUpdater(
                kub_api_service=kas)

            sandbox_deployment_state_tracker = SandboxDeploymentStateTracker(
                logger=logger,
                app_requests=app_requests,
                apps_configuration_end_tracker=apps_configuration_end_tracker,
                sandbox_deployment_end_updater=sandbox_deployment_end_updater,
                sandbox_id=sandbox_id,
                messaging_service=messaging_service,
                space_id=space_id)

            app_instance_identifier_creator = KubAppInstanceIdentifierCreator(kub_api_service=kas)
            status_maintainer = AppStatusMaintainer(logger=logger,
                                                    app_instance_service=app_instance_service,
                                                    apps_configuration_end_tracker=apps_configuration_end_tracker,
                                                    sandbox_deployment_state_tracker=sandbox_deployment_state_tracker,
                                                    sandbox_start_time_updater=sandbox_start_time_updater,
                                                    app_instance_status_event_reporter=app_instance_status_event_reporter)
            return status_maintainer, \
                   app_configuration_start_policy, \
                   app_instance_identifier_creator

        if provider == "aws":
            region_name = config_json['region_name']
            aws_session = AwsSession(region_name=region_name)

            aws_status_maintainer = AWSStatusMaintainer(aws_session, sandbox_id=sandbox_id)

            app_instance_service = AwsAppInstanceService(sandbox_id=sandbox_id,
                                                         logger=logger,
                                                         aws_session=aws_session,
                                                         aws_status_maintainer=aws_status_maintainer)
            apps_configuration_end_tracker = AppsConfigurationEndTracker(logger=logger, app_requests=app_requests,
                                                                         app_instance_service=app_instance_service)
            app_configuration_start_policy = AppConfigurationStartPolicy(apps_configuration_end_tracker,
                                                                         app_requests)
            aws_vpc_service = AwsVpcService(sandbox_id=sandbox_id, aws_session=aws_session)
            sandbox_start_time_updater = AwsSandboxStartTimeUpdater(
                sandbox_id=sandbox_id,
                aws_session=aws_session,
                date_time_provider=DateTimeProvider(),
                logger=logger,
                apps_configuration_end_tracker=apps_configuration_end_tracker,
                vpc_service=aws_vpc_service,
                aws_status_maintainer=aws_status_maintainer
            )

            sandbox_deployment_end_updater = AwsSandboxDeploymentEndUpdater(aws_status_maintainer)

            sandbox_deployment_state_tracker = SandboxDeploymentStateTracker(
                logger=logger,
                app_requests=app_requests,
                apps_configuration_end_tracker=apps_configuration_end_tracker,
                sandbox_deployment_end_updater=sandbox_deployment_end_updater,
                sandbox_id=sandbox_id,
                messaging_service=messaging_service,
                space_id=space_id)

            app_instance_identifier_creator = AwsAppInstanceIdentifierCreator()
            status_maintainer = AppStatusMaintainer(logger=logger,
                                                    app_instance_service=app_instance_service,
                                                    apps_configuration_end_tracker=apps_configuration_end_tracker,
                                                    sandbox_deployment_state_tracker=sandbox_deployment_state_tracker,
                                                    sandbox_start_time_updater=sandbox_start_time_updater,
                                                    app_instance_status_event_reporter=app_instance_status_event_reporter)
            return status_maintainer, \
                   app_configuration_start_policy, \
                   app_instance_identifier_creator

        raise Exception('unknown provider {}'.format(provider))
    except Exception as exc:
        logger.exception(exc)
        raise exc


@app.route("/")
def hello():
    logger.info("start")
    return "welcome cs18:sidecar-api"


@app.route("/application/<string:app_name>/health-check", methods=['POST'])
def start_health_check(app_name: str):
    logger.info(
        "entered with - app_name: {}, request: {}".format(app_name, request.json))
    # noinspection PyBroadException
    try:
        ip_address = request.remote_addr
        req = request.get_json()
        script_name = req['script']
        additional_data = req.get('additional_data')

        logger.info("start_health_check for app '{}', address: {}, request: {}".format(app_name, ip_address, req))
        app_instance_identifier = _app_instance_identifier_creator.create(app_name=app_name,
                                                                          ip_address=ip_address,
                                                                          additional_data=additional_data)
        logger.info("created identifier for '{APP}' on '{IP}': '{APP_ID}'"
                    .format(APP=app_name, IP=ip_address, APP_ID=app_instance_identifier))
        _health_check_monitor.start(ip_address=ip_address, app_instance_identifier=app_instance_identifier,
                                    script_name=script_name)

        # fail here if validation fails ONLY
        return jsonify(success=True), 202
    except Exception as exc:
        logger.exception(exc)
        return jsonify(success=False, error=str(exc)), 401


@app.route('/application/<app_name>/<instance_id>/logs/<topic>', methods=['POST'])
def write_log(app_name: str, instance_id: str, topic: str):
    try:
        events = [(datetime.fromtimestamp(item["date"]), item["line"]) for item in request.json]
        log_entry = LogEntry(app_name, instance_id, topic, events)
        cloud_logger.write(log_entry)

        return jsonify(success=True), 200
    except Exception as exc:
        logger.exception(
            "writing log error, entered with - app_name: {}, instance_id: {}, topic: {}, request: {}, error: {}".format(
                app_name, instance_id, topic,
                request.json, exc))
        return jsonify(success=False, error=str(exc)), 401


@app.route('/application/<string:app_name>/<string:instance_id>/event', methods=['POST'])
def report_app_instance_event(app_name: str, instance_id: str):
    logger.info(
        "entered with - app_name: {}, instance_id: {}, request: {}".format(app_name, instance_id, request.json))
    try:
        data_json = request.get_json()
        event = data_json['event']
        if event not in AppInstanceEvents.ALL:
            error_message = "'{EVENT}' is not one of the known event types".format(EVENT=event)
            return jsonify(success=False, error=error_message), 400

        app_instance_identifier = _app_instance_identifier_creator.create_from_instance_id(app_name=app_name,
                                                                                           instance_id=instance_id)
        _app_instance_event_handler.report_event(app_instance_identifier=app_instance_identifier,
                                                 app_instance_event=event)
        return jsonify(success=True), 200
    except Exception as exc:
        logger.exception(exc)
        return str(exc), 500


@app.route("/application/<string:app_name>/config-start-status", methods=['GET'])
def get_application_configuration_start_status(app_name: str):
    logger.info(
        "entered with - app_name: {}, request: {}".format(app_name, request.json))
    try:
        start_status = _app_configuration_start_policy.get_app_configuration_start_status(app_name)
        return start_status, 200
    except Exception as exc:
        logger.exception(exc)
        return str(exc), 500

def run():
    global _health_check_monitor
    global logger
    global cloud_logger
    global _app_configuration_start_policy
    global _app_instance_identifier_creator
    global _app_instance_event_handler
    # logger definitions
    logger = logging.getLogger('sidecar-api')
    hdlr = logging.FileHandler(Const.get_log_file())
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(filename)s %(funcName)s -> %(message)s')
    hdlr.setFormatter(formatter)
    logger.addHandler(hdlr)
    logger.setLevel(logging.INFO)

    try:
        config_json = _load_config()

        from sidecar.cloud_logger import CloudLoggerFactory
        cloud_logger_factory = CloudLoggerFactory(config_json)
        cloud_logger = cloud_logger_factory.create_instance()

        _app_instance_event_handler = AppInstanceEventHandler(cloud_logger=cloud_logger,
                                                              date_time_provider=DateTimeProvider())

        apps_env = _get_apps_environment_variables(config_json)

        status_maintainer, _app_configuration_start_policy, _app_instance_identifier_creator = \
            get_status_updater(config_json, _app_instance_event_handler)

        _health_check_monitor = HealthCheckMonitor(logger=logger,
                                                   cloud_logger=cloud_logger,
                                                   status_maintainer=status_maintainer,
                                                   apps_env=apps_env)
    except Exception as exc:
        logger.exception(exc)
        raise exc

    # server startstatus_updater
    app.run(host='0.0.0.0', port=4000)


def configure_debug():
    # create log folder: ~/sidecar
    api_log_folder = os.path.dirname(Const.get_log_file())
    file_system = FileSystemService()
    if not file_system.exists(api_log_folder):
        file_system.create_folders(api_log_folder)

    # setup the config file
    # apps_json = {
    #     "myApp": 1
    # }
    apps_json = {
        "demoapp-server": {
            "instances_count": 1,
            "dependencies": [],
             "env":{"WELCOME_STRING": "Welcome to Quali Colony!", "PORT": "3001"},
        }
    }

    messaging_json ={
        "host": "wombat.rmq.cloudamqp.com",
        "user": "rhpjvtoj",
        "password": "M7dZ2D5hWkIEOLPj6poiB4fF1t9cmNIf",
        "queue": "sidecar_queue",
        "exchange": "",
        "routingkey": "sidecar_queue",
        "virtualhost": "rhpjvtoj",
        "port": 5672,
        "queuetype": "fanout"
    }

    config_file_content = {
        "env":"",
        "provider": "aws",
        "sandbox_id": "d229c185-39d5-4d8b-9c4c-1eb4c5d5678a",  # can set it to an existing sandbox
        # "kub_api_address": "https://35.197.122.51",
        "region_name": "eu-west-2",
        "space_id": "84a26312-90d3-4e3e-a0ea-136778623573",
        "apps": apps_json,
        "messaging": messaging_json
    }
    with open(Const.get_config_file(), "w") as conf_file:
        conf_file.write(json.dumps(config_file_content))


if __name__ == "__main__":
    #configure_debug()
    run()



import boto3
from botocore.config import Config


class AwsSession:
    def __init__(self, region_name: str):
        config_dict = {'connect_timeout': 5, 'read_timeout': 15, 'max_pool_connections': 100,
                       'retries': {'max_attempts': 0}}
        self._config = Config(**config_dict)
        self._region_name = region_name
        self.ec2_client = self._get_client('ec2', self._config)
        self.cf_client = self._get_client('cloudformation', self._config)
        self.ec2_resource = self._get_resource('ec2', self._config)
        self._autoscalling_client = self._get_client('autoscaling', self._config)

    def get_ec2_client(self):
        return self.ec2_client

    def get_cf_client(self):
        return self.cf_client

    def get_ec2_resource(self):
        return self.ec2_resource

    def get_autoscaling_client(self):
        return self._autoscalling_client

    def _get_client(self, client: str, config: Config):
        return boto3.client(client, region_name=self._region_name, config=config)

    def _get_resource(self, resource: str, config: Config):
        return boto3.resource(resource, region_name=self._region_name, config=config)

    def get_dynamo_resource(self):
        return boto3.resource('dynamodb', region_name=self._region_name, config=self._config)



import os
import signal
import _thread
from datetime import datetime
from logging import Logger
import subprocess

from sidecar.app_instance_identifier import AppInstanceIdentifier
from sidecar.app_status_maintainer import AppStatusMaintainer
from sidecar.cloud_logger import ICloudLogger, LogEntry
from sidecar.const import Const, AppInstanceConfigStatus
from sidecar.utils import Utils

HEALTH_CHECK_TOPIC = "healthcheck"


class HealthCheckMonitor:
    def __init__(self, logger: Logger, cloud_logger: ICloudLogger,
                 status_maintainer: AppStatusMaintainer, apps_env: dict=None):
        self.logger = logger
        self.cloud_logger = cloud_logger
        self._status_maintainer = status_maintainer
        self.timeout = Const.HEALTH_CHECK_TIMEOUT
        self.apps_env = {} if apps_env is None else apps_env

    def start(self, ip_address: str, app_instance_identifier: AppInstanceIdentifier, script_name: str):
        self.logger.info("entered '{}'".format(app_instance_identifier.name))
        _thread.start_new_thread(self._process_health_check, (ip_address, app_instance_identifier, script_name))

    def _update_status(self, app_instance_identifier: AppInstanceIdentifier, status: str):
        self._status_maintainer.update_status(app_instance_identifier=app_instance_identifier,
                                              status=status)

    def _process_health_check(self, ip_address: str, app_instance_identifier: AppInstanceIdentifier, script_name: str):
        try:
            self._update_status(app_instance_identifier=app_instance_identifier,
                                status=AppInstanceConfigStatus.PENDING)

            result = self._execute_health_check_script(app=app_instance_identifier,
                                                       ip_address=ip_address,
                                                       script_name=script_name)

            status = AppInstanceConfigStatus.COMPLETED if result else AppInstanceConfigStatus.ERROR
            self._update_status(app_instance_identifier=app_instance_identifier,
                                status=status)
        except Exception as exc:
            self.logger.exception(exc)
            raise

    def _execute_health_check_script(self,
                                     app: AppInstanceIdentifier,
                                     ip_address: str,
                                     script_name: str) -> bool:

        self.logger.info("entered '{}'".format(app.name))
        script_file_path = Const.get_health_check_file(app_name=app.name, script_name=script_name)
        command = ['{script_file_path} {ip_address}'.format(
            script_file_path=script_file_path,
            ip_address=ip_address)]

        timestamp = Utils.get_timestamp()
        line = 'health-check started: {0} with command: {1}'.format(timestamp, command)
        log_entry = LogEntry(app.name, app.infra_id, HEALTH_CHECK_TOPIC, [(datetime.utcnow(), line)])
        self.cloud_logger.write(log_entry)

        # get process environment variables and mix them with user defined application specific environment
        # variables
        app_env = self.apps_env.get(app.name)
        env = {**os.environ, **({} if app_env is None else app_env)}

        # run healthcheck command in subprocess and redirect its outputs to subprocess' stdout
        # read stdout line by line until subprocess ended or until timeout and send it to cloud logger
        # if timeout occurred kill healthcheck subprocess

        start = datetime.now()

        with subprocess.Popen(command,
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE,
                              shell=True,
                              preexec_fn=os.setsid,
                              universal_newlines=True,
                              env=env) as p:
            try:
                while True:
                    line = p.stdout.readline()
                    if len(line) == 0 and p.poll() is not None:
                        break

                    log_entry = LogEntry(app.name, app.infra_id, HEALTH_CHECK_TOPIC, [(datetime.utcnow(), line[:-1])])
                    self.cloud_logger.write(log_entry)

                    elapsed = datetime.now() - start
                    if elapsed.total_seconds() > self.timeout:
                        raise subprocess.TimeoutExpired(cmd=command, timeout=self.timeout)

            except subprocess.TimeoutExpired as ex:
                line = 'health-check: timeout after {} seconds'.format(ex.timeout)
                log_entry = LogEntry(app.name, app.infra_id, HEALTH_CHECK_TOPIC, [(datetime.utcnow(), line)])
                self.cloud_logger.write(log_entry)
                self._kill_process(p)

            finally:
                if p.returncode == 0:
                    log_entry = LogEntry(app.name, app.infra_id, HEALTH_CHECK_TOPIC,
                                         [(datetime.utcnow(), 'health-check: done')])
                    self.cloud_logger.write(log_entry)
                    return True
                else:
                    return False

    @staticmethod
    def _kill_process(p: subprocess.Popen):
        try:
            os.killpg(os.getpgid(p.pid), signal.SIGTERM)
        except ProcessLookupError:
            print('process already killed')



import pika
from logging import Logger


class MessagingConnectionProperties:
    def __init__(self, host: str,
                 user: str,
                 password: str,
                 queue: str,
                 exchange: str,
                 routingkey: str,
                 virtualhost: str,
                 port: int,
                 queuetype: str):

        self.host = host
        self.user = user
        self.password = password
        self.queue = queue
        self.exchange = exchange
        self.routingkey = routingkey
        self.virtualhost = virtualhost
        self.port = port
        self.queuetype = queuetype


class MessagingService:
    def __init__(self, connection_props: MessagingConnectionProperties, logger: Logger):
        self._connection_props = connection_props
        self._connection = None
        self._logger = logger


    def publish(self, message: str):

        try:
            credentials = pika.credentials.PlainCredentials(self._connection_props.user,
                                                            self._connection_props.password)

            self._connection = pika.BlockingConnection(pika.ConnectionParameters(self._connection_props.host,
                                                                                 self._connection_props.port,
                                                                                 self._connection_props.virtualhost,
                                                                                 credentials))

            if self._connection is not None:
                channel = self._connection.channel()
                channel.queue_declare(queue=self._connection_props.queue,durable=True)

                channel.basic_publish(exchange=self._connection_props.exchange,
                                      routing_key=self._connection_props.routingkey,
                                      body=message)
                self._closeconnection()
                self._connection = None

        except Exception as exc:
            self._logger.error("an error occurred while connecting to a queue {exc}".format(exc=exc))

        self._closeconnection()
        self._connection = None

    def _closeconnection(self):
        try:
            if self._connection is not None:
                self._connection.close()
        except Exception as exc:
            self._logger.error("an error occurred while closing the RabbitMQ connection {exc}".format(exc=exc))

    # def __exit__(self, exc_type, exc_val,exc_tb):
    #     self.closeconnection()
